@@ -140,7 +140,7 @@ pub fn get_tag(tag: u16, tag_data: &[u8]) -> TagDescription {
 
         0x8104 => tag!(Default, SensorWidth,  "Imager Dimension (Effective Width)",              f32, "{:.2}mm", |d| Ok(d.read_u16::<BigEndian>()? as f32 / 1000.0), tag_data),
         0x8105 => tag!(Default, SensorHeight, "Imager Dimension (Effective Height)",             f32, "{:.2}mm", |d| Ok(d.read_u16::<BigEndian>()? as f32 / 1000.0), tag_data),
-        0x8106 => tag!(Default, FrameRate,    "Capture Frame Rate",                              f32, "{:.3}fps", |d| read_rational(d), tag_data),
+        0x8106 => tag!(Default, FrameRate,    "Capture Frame Rate",                              f64, "{:.3}fps", |d| read_rational(d), tag_data),
         
         0x8107 => tag!(Default, SensorReadoutMode, "Image Sensor Readout Mode", u8, |v| { match v {
             0 => "Interlaced field" .into(),
@@ -167,11 +167,11 @@ pub fn get_tag(tag: u16, tag_data: &[u8]) -> TagDescription {
         0x810E => tag!(Colors, WhiteBalance, "White Balance",                                     u16,  "{}K",    |d| d.read_u16::<BigEndian>(), tag_data),
         0x810F => tag!(Colors, MasterBlackLevel, "Camera Master BlackLevel",                      f32,  "{:.2}",  |d| Ok(d.read_u16::<BigEndian>()? as f32 / 10.0), tag_data),
         0x8110 => tag!(Colors, KneePoint, "Camera Knee Point",                                    f32,  "{:.2}",  |d| Ok(d.read_u16::<BigEndian>()? as f32 / 10.0), tag_data),
-        0x8111 => tag!(Colors, KneeSlope, "Camera Knee Slope",                                    f32,  "{:.2}",  |d| read_rational(d), tag_data),
+        0x8111 => tag!(Colors, KneeSlope, "Camera Knee Slope",                                    f64,  "{:.2}",  |d| read_rational(d), tag_data),
         0x8112 => tag!(Colors, LuminanceDynamicRange, "Camera Luminance Dynamic Range",           f32,  "{:.2}",  |d| Ok(d.read_u16::<BigEndian>()? as f32 / 10.0), tag_data),
-        0x8113 => tag!(Default, TagId::Custom("SettingFileURI".into()), "Camera Setting File URI",       String, |v| v.to_string(), |d| read_utf8(d), tag_data),
+        0x8113 => tag!(Default, TagId::Custom("SettingFileURI".into()), "Camera Setting File URI",String, |v| v.to_string(), |d| read_utf8(d), tag_data),
         0x8114 => tag!(Default, CameraAttributes, "Camera Attributes",                            String, |v| v.to_string(), |d| read_utf8(d), tag_data),
-        0x8115 => tag!(Exposure, TagId::Custom("ISOValue2".into()), "Exposure Index of Photo Meter",     u16, "{}",      |d| d.read_u16::<BigEndian>(), tag_data),
+        0x8115 => tag!(Exposure, TagId::Custom("ISOValue2".into()), "Exposure Index of Photo Meter", u16, "{}",   |d| d.read_u16::<BigEndian>(), tag_data),
         
         0x8116 => tag!(Colors, TagId::Custom("GammaforCDL".into()), "Gamma for CDL", u8, |v| { match v {
             0 => "Same as Capture Gamma".into(),
@@ -182,8 +182,7 @@ pub fn get_tag(tag: u16, tag_data: &[u8]) -> TagDescription {
             _ => format!("{}", v)
         } }, |d| d.read_u8(), tag_data),
 
-        0x8117 => tag!(Colors, TagId::Custom("ASCCDLValue".into()), "ASC CDL V1.2", String, |v| v.to_string(), |d| {
-            // TODO: Make separate type instead of string
+        0x8117 => tag!(Colors, TagId::Custom("ASCCDLValue".into()), "ASC CDL V1.2", Json, |v| v.to_string(), |d| {
             let count = d.read_u32::<BigEndian>()?;
             let length = d.read_u32::<BigEndian>()?;
             if count != 10 || length != 2 { return Err(Error::new(ErrorKind::Other, "Invalid")); }
@@ -192,11 +191,15 @@ pub fn get_tag(tag: u16, tag_data: &[u8]) -> TagDescription {
             let pr = read_f16_corrected(d)?; let pg = read_f16_corrected(d)?; let pb = read_f16_corrected(d)?;
 
             let sat = read_f16_corrected(d)?;
-            Ok(format!("sR={:.1} sG={:.1} sB={:.1}\noR={:.1} oG={:.1} oB={:.1}\npR={:.1} pG={:.1} pB={:.1}\nsat={:.1}", sr, sg, sb, or, og, ob, pr, pg, pb, sat))
+            Ok(serde_json::json!({
+                "slope":  { "r": sr, "g": sg, "b": sb},
+                "offset": { "r": or, "g": og, "b": ob},
+                "power":  { "r": pr, "g": pg, "b": pb},
+                "saturation": sat
+            }))
         }, tag_data),
 
-        0x8118 => tag!(Colors, ColorMatrix, "Color matrix", String, |v| v.to_string(), |d| {
-            // TODO: Make separate type instead of string
+        0x8118 => tag!(Colors, ColorMatrix, "Color matrix", Json, |v| v.to_string(), |d| {
             let count  = d.read_u32::<BigEndian>()?;
             let length = d.read_u32::<BigEndian>()?;
             if count != 9 || length != 8 {
@@ -214,11 +217,112 @@ pub fn get_tag(tag: u16, tag_data: &[u8]) -> TagDescription {
             let rb = d.read_u32::<BigEndian>()? as f32 / d.read_u32::<BigEndian>()? as f32;
             let gb = d.read_u32::<BigEndian>()? as f32 / d.read_u32::<BigEndian>()? as f32;
             let bb = d.read_u32::<BigEndian>()? as f32 / d.read_u32::<BigEndian>()? as f32;
-            Ok(format!("RR={:.3} GR={:.3} BR={:.3}\nRG={:.3} GG={:.3} BG={:.3}\nRB={:.3} GB={:.3} BB={:.3}", rr, gr, br, rg, gg, bg, rb, gb, bb))
+            Ok(serde_json::json!({
+                "RR": rr, "GR": gr, "BR": br,
+                "RG": rg, "GG": gg, "BG": bg,
+                "RB": rb, "GB": gb, "BB": bb
+            }))
         }, tag_data),
 
         // -------------- UserDefinedAcquisitionMetadata --------------
         0xe000 => tag!(Default, GroupIdentifier, "UDAM Set Identifier", Uuid, |v| format!("{{{:08x}-{:08x}-{:08x}-{:08x}}}", v.0, v.1, v.2, v.3), |d| read_uuid(d), tag_data),
+        
+        0xe101 => tag!(Default, TagId::Custom("EffectiveMarkerCoverage".into()),         "Effective marker coverage",       u32x2, |v| format!("{} x {}", v.0, v.1), |d| Ok((d.read_u32::<BigEndian>()?, d.read_u32::<BigEndian>()?)), tag_data),
+        0xe102 => tag!(Default, TagId::Custom("EffectiveMarkerAspectRatio".into()),      "Effective marker aspect ratio",   u32x2, |v| format!("{} x {}", v.0, v.1), |d| Ok((d.read_u32::<BigEndian>()?, d.read_u32::<BigEndian>()?)), tag_data),
+        0xe103 => tag!(Default, TagId::Custom("CameraProcessDiscriminationCode".into()), "Camera process discrimination code", u16,|v| { match v {
+            0x0101 => "F65 RAW Mode released in December 2011".into(),
+            0x0102 => "F65 HD Mode released in April 2012".into(),
+            0x0103 => "F65 RAW High Frame Rate Mode released in July 2012".into(),
+            _ => format!("0x{:04x}", v)
+        } }, |d| d.read_u16::<BigEndian>(), tag_data),
+        0xe104 => tag!(Default, TagId::Custom("RotaryShutterMode".into()),               "Rotary shutter mode",             bool,   "{}", |d| Ok(d.read_u8()? != 0), tag_data), 
+        0xe105 => tag!(Default, TagId::Custom("RawBlackCodeValue".into()),               "RawBlack code value",             u16,    "{}", |d| d.read_u16::<BigEndian>(), tag_data), 
+        0xe106 => tag!(Default, TagId::Custom("RawGrayCodeValue".into()),                "RawGray code value",              u16,    "{}", |d| d.read_u16::<BigEndian>(), tag_data), 
+        0xe107 => tag!(Default, TagId::Custom("RawWhiteCodeValue".into()),               "RawWhite code value",             u16,    "{}", |d| d.read_u16::<BigEndian>(), tag_data), 
+        0xe109 => tag!(Default, TagId::Custom("MonitoringDescriptions".into()),          "Monitoring descriptions",         String, "{}", |d| read_utf8(d), tag_data), 
+        0xe10B => tag!(Default, TagId::Custom("MonitoringBaseCurve".into()),             "Monitoring base curve",           Uuid, |v| { match v.3 {
+            0x01010000 => "BT.470"                    .into(),
+            0x01020000 => "BT.709"                    .into(),
+            0x01030000 => "SMPTE ST 240"              .into(),
+            0x01040000 => "SMPTE ST 274"              .into(),
+            0x01050000 => "BT.1361"                   .into(),
+            0x01060000 => "SceneLinear"               .into(),
+            0x01080000 => "Rec709-xvYCC"              .into(),
+            0x010b0000 => "Rec2100-HLG"               .into(),
+            0x01010101 => "DVW-709 Like"              .into(),
+            0x01010102 => "E10/E30STD for J EK"       .into(),
+            0x01010103 => "E10/E30STD for UC"         .into(),
+            0x01010106 => "BBC Initial50"             .into(),
+            0x01010107 => "SD CamCorder STD"          .into(),
+            0x01010108 => "BVW-400 Like"              .into(),
+            0x01010109 => "Ikegami"                   .into(),
+            0x0101017F => "reproduced unknown label"  .into(),
+            0x01010201 => "HG3250G36"                 .into(),
+            0x01010202 => "HG4600G30"                 .into(),
+            0x01010203 => "HG3259G40"                 .into(),
+            0x01010204 => "HG4609G33"                 .into(),
+            0x01010205 => "HG8000G36"                 .into(),
+            0x01010206 => "HG8000G30"                 .into(),
+            0x01010207 => "HG8009G40"                 .into(),
+            0x01010208 => "HG8009G33"                 .into(),
+            0x01010301 => "CINE1 of EX1/EX3"          .into(),
+            0x01010302 => "CINE2 of EX1/EX3"          .into(),
+            0x01010303 => "CINE3 of EX1/EX3"          .into(),
+            0x01010304 => "CINE4 of EX1/EX3"          .into(),
+            0x01010305 => "Kodak 5248 film like"      .into(),
+            0x01010306 => "Kodak 5245 film like"      .into(),
+            0x01010307 => "Kodak 5293 film like"      .into(),
+            0x01010308 => "Kodak 5296 film like"      .into(),
+            0x01010309 => "Average of Film of MSW-900".into(),
+            0x01010401 => "User defined curve1"       .into(),
+            0x01010402 => "User defined curve2"       .into(),
+            0x01010403 => "User defined curve3"       .into(),
+            0x01010404 => "User defined curve4"       .into(),
+            0x01010405 => "User defined curve5"       .into(),
+            0x01010406 => "User defined curve6"       .into(),
+            0x01010407 => "User defined curve7"       .into(),
+            0x01010408 => "User defined curve8"       .into(),
+            0x01010501 => "S-Log"                     .into(),
+            0x01010502 => "FS-Log"                    .into(),
+            0x01010503 => "R709 180%"                 .into(),
+            0x01010504 => "R709 800%"                 .into(),
+            0x01010506 => "Cine-Log"                  .into(),
+            0x01010507 => "ASC-CDL"                   .into(),
+            0x01010508 => "S-Log2"                    .into(),
+            0x01010602 => "Still"                     .into(),
+            0x01010604 => "S-Log3"                    .into(),
+            0x01010605 => "S-Log3-Cine"               .into(),
+            _ => format!("{{{:08x}-{:08x}-{:08x}-{:08x}}}", v.0, v.1, v.2, v.3)
+        } }, |d| read_uuid(d), tag_data),
+        /*0xe201 => tag!(Default, TagId::Custom("CookeProtocol_BinaryMetadata".into()),    "CookeProtocol_BinaryMetadata",    Json, "{:?}", |d| {
+            // TODO: Implement parsing of this type
+            // https://github.com/MediaArea/MediaInfoLib/blob/master/Source/MediaInfo/Multiple/File_Mxf.cpp#L12584 - UserDefinedAcquisitionMetadata_Sony_E201
+            "FocusDistance",
+            "ApertureValue",
+            "ApertureScale",
+            "EffectiveFocaleLength",
+            "HyperfocalDistance",
+            "NearFocusDistance",
+            "FarFocusDistance",
+            "HorizontalFieldOfView",
+            "EntrancePupilPosition",
+            "NormalizedZoomValue",
+            "LensSerialNumber",
+        }, tag_data), */
+        0xe202 => tag!(Default, TagId::Custom("CookeProtocol_UserMetadata".into()),    "CookeProtocol_UserMetadata",      String, "{}", |d| read_utf8(d), tag_data), 
+        0xe203 => tag!(Default, TagId::Custom("CookeProtocol_CalibrationType".into()), "CookeProtocol_CalibrationType",   u8, |v| { match v {
+            0 => "mm".into(),
+            1 => "in".into(),
+            _ => format!("{}", v)
+        } }, |d| d.read_u8(), tag_data),
+        0xe108 => tag!(Default, Unknown(0xe108), "Unknown_e108", Uuid, |v| format!("{{{:08x}-{:08x}-{:08x}-{:08x}}}", v.0, v.1, v.2, v.3), |d| read_uuid(d), tag_data),
+        0xe10d => tag!(Default, Unknown(0xe10d), "Unknown_e10d", Uuid, |v| format!("{{{:08x}-{:08x}-{:08x}-{:08x}}}", v.0, v.1, v.2, v.3), |d| read_uuid(d), tag_data),
+        0xe10e => tag!(Default, Unknown(0xe10e), "Unknown_e10e", Uuid, |v| format!("{{{:08x}-{:08x}-{:08x}-{:08x}}}", v.0, v.1, v.2, v.3), |d| read_uuid(d), tag_data),
+        0xe10f => tag!(Default, Unknown(0xe10f), "Unknown_e10f", u32x2, |v| format!("{} x {}", v.0, v.1), |d| Ok((d.read_u32::<BigEndian>()?, d.read_u32::<BigEndian>()?)), tag_data),
+        
+        0xe111 => tag!(Default, Unknown(0xe111), "Unknown_e111", Uuid, |v| format!("{{{:08x}-{:08x}-{:08x}-{:08x}}}", v.0, v.1, v.2, v.3), |d| read_uuid(d), tag_data),
+        0xe112 => tag!(Default, Unknown(0xe112), "Unknown_e112", Uuid, |v| format!("{{{:08x}-{:08x}-{:08x}-{:08x}}}", v.0, v.1, v.2, v.3), |d| read_uuid(d), tag_data),
+        0xe113 => tag!(Default, Unknown(0xe113), "Unknown_e113", String, "{}", |d| read_utf8(d), tag_data),
 
         // -------------- Sony's proprietary --------------
         0xe300 => tag!(Default, StabilizationEnabled, "Stabilization", u8, "{}", |d| d.read_u8(), tag_data),
@@ -243,23 +347,22 @@ pub fn get_tag(tag: u16, tag_data: &[u8]) -> TagDescription {
             Ok(chrono::NaiveDate::from_ymd((yy1 * 100.0 + yy2) as i32, mm, dd).and_hms(h, m, s).timestamp() as u64)
         }, tag_data),
 
-
         // Possible values: zFar, zNear, aspect, temporal_position, temporal_rotation
         ////////////////////////////////////////// ImagerControlInformation (IBIS) //////////////////////////////////////////
-        0xe400 => tag!(IBIS, Unknown(tag as u32), "IBIS position/rotation 3xi32", String, |v| v.to_string(), |d| {
+        0xe400 => tag!(IBIS, Unknown(tag as u32), "IBIS position/rotation 3xi32", Vector3_i32, "{:?}", |d| {
             let x = d.read_i32::<BigEndian>()?;
             let y = d.read_i32::<BigEndian>()?;
             let z = d.read_i32::<BigEndian>()?;
-            Ok(format!("{} {} {}", x, y, z))
+            Ok(Vector3 { x, y, z })
         }, tag_data),
         0xe401 => tag!(IBIS, Unknown(tag as u32), "IBIS position/rotation u8", u8, "{}", |d| d.read_u8(), tag_data),
         0xe402 => tag!(IBIS, Unknown(tag as u32), "IBIS position/rotation i32", i32, "{}", |d| d.read_i32::<BigEndian>(), tag_data),
         0xe403 => tag!(IBIS, Unknown(tag as u32), "IBIS position/rotation u8", u8, "{}", |d| d.read_u8(), tag_data),
-        0xe404 => tag!(IBIS, Unknown(tag as u32), "IBIS Position/Rotation 3xi16", String, |v| v.to_string(), |d| {
+        0xe404 => tag!(IBIS, Unknown(tag as u32), "IBIS Position/Rotation 3xi16", Vector3_i16, "{:?}", |d| {
             let x = d.read_i16::<BigEndian>()?;
             let y = d.read_i16::<BigEndian>()?;
             let z = d.read_i16::<BigEndian>()?;
-            Ok(format!("{} {} {}", x, y, z))
+            Ok(Vector3 { x, y, z })
         }, tag_data),
         0xe405 => tag!(IBIS, Unknown(tag as u32), "IBIS 2xi16", String, |v| v.to_string(), |d| {
             let x = d.read_i16::<BigEndian>()?;
@@ -293,17 +396,21 @@ pub fn get_tag(tag: u16, tag_data: &[u8]) -> TagDescription {
             if length != 16 {
                 return Err(Error::new(ErrorKind::Other, "Invalid OSS table"));
             }
-            let mut ret = Vec::with_capacity(count as usize);
-            for _ in 0..count {
-                // XAVC::base_2D_TimeOffset<XAVC::base_3D<int>>
-                ret.push(TimeVector3 {
-                    t: d.read_i32::<BigEndian>()?, // time offset
-                    x: d.read_i32::<BigEndian>()?, // x, confirmed i32
-                    y: d.read_i32::<BigEndian>()?, // y, confirmed i32
-                    z: d.read_i32::<BigEndian>()?  // z. confirmed i32
-                });
+            if count > 0 {
+                let mut ret = Vec::with_capacity(count as usize);
+                for _ in 0..count {
+                    // XAVC::base_2D_TimeOffset<XAVC::base_3D<int>>
+                    ret.push(TimeVector3 {
+                        t: d.read_i32::<BigEndian>()?, // time offset
+                        x: d.read_i32::<BigEndian>()?, // x, confirmed i32
+                        y: d.read_i32::<BigEndian>()?, // y, confirmed i32
+                        z: d.read_i32::<BigEndian>()?  // z. confirmed i32
+                    });
+                }
+                Ok(ret)
+            } else {
+                Ok(vec![])
             }
-            Ok(ret)
         }, tag_data),
         0xe450 => tag!(IBIS, Data2, "IBIS TimeOffset table 2", Vec_TimeVector3_i32, "{:?}", |d| {
             let count  = d.read_i32::<BigEndian>()?;
@@ -311,17 +418,21 @@ pub fn get_tag(tag: u16, tag_data: &[u8]) -> TagDescription {
             if length != 10 {
                 return Err(Error::new(ErrorKind::Other, "Invalid table"));
             }
-            let mut ret = Vec::with_capacity(count as usize);
-            for _ in 0..count {
-                // XAVC::base_2D_TimeOffset<XAVC::base_3D<short>>
-                ret.push(TimeVector3 {
-                    t: d.read_i32::<BigEndian>()?, // time offset
-                    x: d.read_i16::<BigEndian>()? as i32, // x, confirmed i16
-                    y: d.read_i16::<BigEndian>()? as i32, // y, confirmed i16
-                    z: d.read_i16::<BigEndian>()? as i32  // z, confirmed i16
-                });
+            if count > 0 {
+                let mut ret = Vec::with_capacity(count as usize);
+                for _ in 0..count {
+                    // XAVC::base_2D_TimeOffset<XAVC::base_3D<short>>
+                    ret.push(TimeVector3 {
+                        t: d.read_i32::<BigEndian>()?, // time offset
+                        x: d.read_i16::<BigEndian>()? as i32, // x, confirmed i16
+                        y: d.read_i16::<BigEndian>()? as i32, // y, confirmed i16
+                        z: d.read_i16::<BigEndian>()? as i32  // z, confirmed i16
+                    });
+                }
+                Ok(ret)
+            } else {
+                Ok(vec![])
             }
-            Ok(ret)
         }, tag_data),
         ////////////////////////////////////////// ImagerControlInformation (IBIS) //////////////////////////////////////////
 
@@ -349,38 +460,47 @@ pub fn get_tag(tag: u16, tag_data: &[u8]) -> TagDescription {
             if length != 16 {
                 return Err(Error::new(ErrorKind::Other, "Invalid table"));
             }
-            let mut ret = Vec::with_capacity(count as usize);
-            for _ in 0..count {
-                // XAVC::base_2D_TimeOffset<XAVC::base_3D<int>>
-                ret.push(TimeVector3 {
-                    t: d.read_i32::<BigEndian>()?, // time offset
-                    x: d.read_i32::<BigEndian>()?, // x
-                    y: d.read_i32::<BigEndian>()?, // y
-                    z: d.read_i32::<BigEndian>()?  // z
-                });
+            if count > 0 {
+                let mut ret = Vec::with_capacity(count as usize);
+                for _ in 0..count {
+                    // XAVC::base_2D_TimeOffset<XAVC::base_3D<int>>
+                    ret.push(TimeVector3 {
+                        t: d.read_i32::<BigEndian>()?, // time offset
+                        x: d.read_i32::<BigEndian>()?, // x
+                        y: d.read_i32::<BigEndian>()?, // y
+                        z: d.read_i32::<BigEndian>()?  // z
+                    });
+                }
+                Ok(ret)
+            } else {
+                Ok(vec![])
             }
-            Ok(ret)
         }, tag_data),
         ////////////////////////////////////////// LensControlInformation (Lens OSS) //////////////////////////////////////////
 
         ////////////////////////////////////////// DistortionCorrection //////////////////////////////////////////
         0xe420 => tag!(GroupId::Custom("LensDistortion".into()), Enabled, "LensDistortion bool", bool, "{}", |d| Ok(d.read_u8()? != 0), tag_data),
-        0xe421 => tag!(GroupId::Custom("LensDistortion".into()), Data,    "LensDistortion Table", String, |v| v.to_string(), |d| {
+        0xe421 => tag!(GroupId::Custom("LensDistortion".into()), Data,    "LensDistortion Table", Json, |v| v.to_string(), |d| {
             let aa = d.read_u32::<BigEndian>()?; // confirmed u32
             let bb = d.read_u32::<BigEndian>()?; // confirmed u32
 
             let cc = d.read_u8()?; // confirmed u8
-            let dd = d.read_f32::<BigEndian>()?; // confirmed u32
+            let dd = d.read_f32::<BigEndian>()?; // confirmed f32
             let elem_count = d.read_u32::<BigEndian>()?;
             let _elem_size = d.read_u32::<BigEndian>()?;
             let mut ret = Vec::with_capacity(elem_count as usize); // &XAVC::base_Array<unsigned short>
             for _ in 0..elem_count {
                 ret.push(d.read_u16::<BigEndian>()?); // confirmed u16
             }
-            Ok(format!("{:?} {} {} {:?}", (aa, bb), cc, dd, ret))
+            Ok(serde_json::json!({
+                "unk1": [aa, bb],
+                "unk2": cc,
+                "unk3": dd,
+                "unk4": ret
+            }))
         }, tag_data),
         0xe422 => tag!(GroupId::Custom("FocalPlaneDistortion".into()), Enabled, "FocalPlaneDistortion bool", bool, "{}", |d| Ok(d.read_u8()? != 0), tag_data),
-        0xe423 => tag!(GroupId::Custom("FocalPlaneDistortion".into()), Data,    "FocalPlaneDistortion Table", String, |v| v.into(), |d| {
+        0xe423 => tag!(GroupId::Custom("FocalPlaneDistortion".into()), Data,    "FocalPlaneDistortion Table", Json, |v| v.to_string(), |d| {
             let aa = d.read_i32::<BigEndian>()?;
             let bb = d.read_i16::<BigEndian>()?;
             let cc = d.read_i16::<BigEndian>()?;
@@ -393,10 +513,15 @@ pub fn get_tag(tag: u16, tag_data: &[u8]) -> TagDescription {
                     d.read_i16::<BigEndian>()?, // y
                 ));
             }
-            Ok(format!("{} {} {} {:?}", aa, bb, cc, ret))
+            Ok(serde_json::json!({
+                "unk1": aa,
+                "unk2": bb,
+                "unk3": cc,
+                "unk4": ret
+            }))
         }, tag_data),
         0xe424 => tag!(GroupId::Custom("MeshCorrection".into()), Enabled, "MeshCorrection::Mesh bool", bool, "{}", |d| Ok(d.read_u8()? != 0), tag_data),
-        0xe42f => tag!(GroupId::Custom("MeshCorrection".into()), Data,    "MeshCorrection::Mesh", String, |v| v.to_string(), |x| {
+        0xe42f => tag!(GroupId::Custom("MeshCorrection".into()), Data,    "MeshCorrection::Mesh", Json, |v| v.to_string(), |x| {
             let aa = x.read_i16::<BigEndian>()?; // confirmed i16
             
             let bb = x.read_u32::<BigEndian>()?; // confirmed i32
@@ -415,7 +540,13 @@ pub fn get_tag(tag: u16, tag_data: &[u8]) -> TagDescription {
             let f3 = x.read_u8()?;
             let f4 = x.read_u8()?;
 
-            Ok(format!("{}, {:?}, {:?}\nXs: {:?}\nYs: {:?}\nSize: {} {} {} {}", aa, (bb, cc), (dd, ee), xs, ys, f1, f2, f3, f4))
+            Ok(serde_json::json!({
+                "unk1": aa,
+                "unk2": [bb, cc],
+                "size": [dd, ee],
+                "points": xs.into_iter().zip(ys).collect::<Vec<_>>(),
+                "unk3": [f1, f2, f3, f4]
+            }))
         }, tag_data),
         ////////////////////////////////////////// DistortionCorrection //////////////////////////////////////////
 
@@ -446,20 +577,24 @@ pub fn get_tag(tag: u16, tag_data: &[u8]) -> TagDescription {
         0xe439 => tag!(Gyroscope, Scale,           "Gyroscope scale", f32, "{}", |d| d.read_f32::<BigEndian>(), tag_data),
         0xe43a => tag!(Gyroscope, Orientation,     "Gyroscope orientation", String, "{}", read_orientation, tag_data),
         0xe43b => tag!(Gyroscope, Data,            "Gyroscope data", Vec_Vector3_i16, "{:?}", |d| {
-            let count = d.read_u32::<BigEndian>()?;
-            let length = d.read_u32::<BigEndian>()?;
+            let count = d.read_i32::<BigEndian>()?;
+            let length = d.read_i32::<BigEndian>()?;
             if length != 6 {
                 return Err(Error::new(ErrorKind::Other, "Invalid gyro data format"));
             }
-            let mut ret = Vec::with_capacity(count as usize);
-            for _ in 0..count {
-                ret.push(Vector3 {
-                    x: d.read_i16::<BigEndian>()?, // pitch
-                    y: d.read_i16::<BigEndian>()?, // roll
-                    z: d.read_i16::<BigEndian>()?, // yaw
-                });
+            if count > 0 {
+                let mut ret = Vec::with_capacity(count as usize);
+                for _ in 0..count {
+                    ret.push(Vector3 {
+                        x: d.read_i16::<BigEndian>()?, // pitch
+                        y: d.read_i16::<BigEndian>()?, // roll
+                        z: d.read_i16::<BigEndian>()?, // yaw
+                    });
+                }
+                Ok(ret)
+            } else {
+                Ok(vec![])
             }
-            Ok(ret)
         }, tag_data),
         ////////////////////////////////////////// Gyroscope //////////////////////////////////////////
         ////////////////////////////////////////// Accelerometer //////////////////////////////////////////
@@ -495,40 +630,72 @@ pub fn get_tag(tag: u16, tag_data: &[u8]) -> TagDescription {
             if length != 6 {
                 return Err(Error::new(ErrorKind::Other, "Invalid accel data format"));
             }
-            let mut ret = Vec::with_capacity(count as usize);
-            for _ in 0..count {
-                ret.push(Vector3 {
-                    x: d.read_i16::<BigEndian>()?, // X
-                    y: d.read_i16::<BigEndian>()?, // Y
-                    z: d.read_i16::<BigEndian>()?, // Z
-                });
+            if count > 0 {
+                let mut ret = Vec::with_capacity(count as usize);
+                for _ in 0..count {
+                    ret.push(Vector3 {
+                        x: d.read_i16::<BigEndian>()?, // X
+                        y: d.read_i16::<BigEndian>()?, // Y
+                        z: d.read_i16::<BigEndian>()?, // Z
+                    });
+                }
+                Ok(ret)
+            } else {
+                Ok(vec![])
             }
-            Ok(ret)
         }, tag_data),
         ////////////////////////////////////////// Accelerometer //////////////////////////////////////////
 
         0xf010 => tag!(UnknownGroup(0xf000), Unknown(tag as u32), "Large unknown", tag_data),
         0xf020 => tag!(UnknownGroup(0xf000), Unknown(tag as u32), "Large unknown", tag_data),
 
-        /* TODO: GPS Tags
-           0x8500 - GPSVersionID 4 bytes - gps version - 2.2.0.0 (02020000)
-           0x8501 - GPSLatitudeRef 1 byte - LatitudeRef - N (4e)
-           0x8502 - GPSLatitude 18h bytes - Latitude - [4]/[4]:[4]/[4]:[4]/[4] = 09:09:09.123
-           0x8503 - GPSLongitudeRef 1 byte - LongtitudeRef - E (45)
-           0x8504 - GPSLongitude 18h bytes - Longtitude - [4]/[4]:[4]/[4]:[4]/[4] = 09:09:09.123
-           0x8505 - 1 byte - AltitudeRef  (equal to 1)
-           0x8506 - 8 bytes - Altitude (meters) ([4]/[4]???). Second [4] almost always = 1000 dec
-           0x8507 - GPSTimeStamp 18h bytes - Timestamp - [4]/[4]:[4]/[4]:[4]/[4] = 09:09:09.123
-           0x8509 - GPSStatus 1 byte - STATUS - 'A' (if GPS not acquired, = 'V')
-           0x850a - GPSMeasureMode 1 byte - MeasureMode - (2 = 2D, 3 = 3D)
-           0x850b - 8 bytes - DOP ([4]/[4]???). Second [4] almost always = 1000 dec
-           0x850c - 1 byte - SpeedRef (K = km/h, M = mph, N = knots)
-           0x850d - 8 bytes ([4]/[4]???) - SPEED
-           0x850e - 1 byte - TrackRef (Direction Reference, T = True direction, M = Magnetic direction)
-           0x850f - Direction 8 bytes ([4]/[4]???) (degrees from 0.0 to 359.99)
-           0x8512 - GPSMapDatum 6 bytes - MapDatum  - 57 47 53 2D 38 34 (WGS-84)
-           0x851d - GPSDateStamp 0a bytes - string (2018:10:30)
-        */
+        ////////////////////////////////////////// GPS //////////////////////////////////////////
+        0x8500 => tag!(GPS, TagId::Custom("GPSVersionID".into()), "GPS version ID", u32, |v| format!("{}.{}.{}.{}", (v << 24) & 0xff, (v << 16) & 0xff, (v << 8) & 0xff, v & 0xff), |d| d.read_u32::<BigEndian>(), tag_data),
+        0x8501 => tag!(GPS, TagId::Custom("GPSLatitudeRef".into()), "GPS latitude ref", u8, |v| match v {
+            b'N' => "North".into(),
+            b'S' => "South".into(),
+            _ => format!("{}", *v as char)
+        }, |d| d.read_u8(), tag_data),
+        0x8502 => tag!(GPS, TagId::Custom("GPSLatitude".into()), "GPS latitude", f64x3, |v| format!("{}:{}:{}", v.0, v.1, v.2), |d| Ok((read_rational(d)?, read_rational(d)?, read_rational(d)?)), tag_data), // TODO: ToDegrees, ToDMS
+        0x8503 => tag!(GPS, TagId::Custom("GPSLongitudeRef".into()), "GPS longitude ref", u8, |v| match v {
+            b'E' => "East".into(),
+            b'W' => "West".into(),
+            _ => format!("{}", *v as char)
+        }, |d| d.read_u8(), tag_data),
+        0x8504 => tag!(GPS, TagId::Custom("GPSLongitude".into()), "GPS longitude", f64x3, |v| format!("{}:{}:{}", v.0, v.1, v.2), |d| Ok((read_rational(d)?, read_rational(d)?, read_rational(d)?)), tag_data), // TODO: ToDegrees, ToDMS
+        0x8505 => tag!(GPS, TagId::Custom("GPSAltitudeRef".into()), "GPS altitude ref", u8, "{}", |d| d.read_u8(), tag_data),
+        0x8506 => tag!(GPS, TagId::Custom("GPSAltitude".into()), "GPS altitude", f64, "{}", |d| read_rational(d), tag_data),
+        
+        0x8507 => tag!(GPS, TagId::Custom("GPSTimeStamp".into()), "GPS timestamp", f64x3, |v| format!("{}:{}:{}", v.0, v.1, v.2), |d| Ok((read_rational(d)?, read_rational(d)?, read_rational(d)?)), tag_data), // TODO: ConvertTimeStamp, PrintTimeStamp
+        0x8509 => tag!(GPS, TagId::Custom("GPSStatus".into()), "GPS status", u8, |v| match v {
+            b'A' => "Active".into(),
+            b'V' => "Void".into(),
+            _ => format!("{}", *v as char)
+        }, |d| d.read_u8(), tag_data),
+        0x850a => tag!(GPS, TagId::Custom("GPSMeasureMode".into()), "GPS measure mode", u8, |v| match v {
+            2 => "2D".into(),
+            3 => "3D".into(),
+            _ => format!("{}", v)
+        }, |d| d.read_u8(), tag_data),
+
+        0x850b => tag!(GPS, TagId::Custom("DOP".into()), "DOP", f64, "{}", |d| read_rational(d), tag_data),
+        0x850c => tag!(GPS, TagId::Custom("GPSSpeedRef".into()), "GPS speed ref", u8, |v| match v {
+            b'K' => "km/h".into(),
+            b'M' => "mph".into(),
+            b'N' => "knots".into(),
+            _ => format!("{}", *v as char)
+        }, |d| d.read_u8(), tag_data),
+        0x850d => tag!(GPS, TagId::Custom("GPSSpeed".into()), "GPS speed", f64, "{}", |d| read_rational(d), tag_data),
+        0x850e => tag!(GPS, TagId::Custom("GPSTrackRef".into()), "GPS track ref", u8, |v| match v {
+            b'T' => "True direction".into(),
+            b'M' => "Magnetic direction".into(),
+            _ => format!("{}", *v as char)
+        }, |d| d.read_u8(), tag_data),
+
+        0x8512 => tag!(GPS, TagId::Custom("GPSMapDatum".into()),  "GPS map datum",  String, |v| v.to_string(), |d| read_utf8(d), tag_data),
+        0x851d => tag!(GPS, TagId::Custom("GPSDateStamp".into()), "GPS date stamp", String, |v| v.to_string(), |d| read_utf8(d), tag_data), // TODO: Exif::ExifDate
+
+        ////////////////////////////////////////// GPS //////////////////////////////////////////
         _ => tag!(UnknownGroup(0), Unknown(tag as u32), "Unknown", tag_data),
     }
 }
@@ -569,47 +736,33 @@ fn read_uuid(d: &mut Cursor::<&[u8]>) -> Result<(u32,u32,u32,u32)> {
 
 fn read_orientation(d: &mut Cursor::<&[u8]>) -> Result<String> {
     let num = d.read_u16::<BigEndian>()?;
-    // RX0 II:    0x241 ; 0010 0100 0001 ; xZY -> Zxy
-    // A7s III:   0x420 ; 0100 0010 0000 ; XYZ -> YXz
-    // RX100 VII: 0x152 ; 0001 0101 0010 ; Yzx -> zYX
-    // ZV1:       0x351 ; 0011 0101 0001 ; xzy -> zxY
-    // A7C:       0x420 ; 0100 0010 0000 ; XYZ -> YXz
+    // RX0 II:    0x241 ; 0010 0100 0001 ; xZY
+    // A7s III:   0x420 ; 0100 0010 0000 ; XYZ
+    // RX100 VII: 0x152 ; 0001 0101 0010 ; Yzx
+    // ZV1:       0x351 ; 0011 0101 0001 ; xzy
+    // A7C:       0x420 ; 0100 0010 0000 ; XYZ
 
-    fn from_num(n: i8) -> char {
-        match n { // lowercase is negative
+    fn from_num(n: i8) -> Result<char> {
+        Ok(match n { // lowercase is negative
             0 => 'X', 1 => 'x',
             2 => 'Y', 3 => 'y',
             4 => 'Z', 5 => 'z',
-            _ => '_'
-        }
-    }
-    fn invert_case(x: char) -> char {
-        if x.is_ascii_lowercase() { x.to_ascii_uppercase() } else { x.to_ascii_lowercase() }
+            _ => { return Err(Error::new(ErrorKind::Other, format!("Invalid orientation data! {} {:#x} {:#b}", n, n, n))); }
+        })
     }
 
-    let mut ret = vec![
-        from_num((num & 0x0f) as i8),
-        from_num(((num >> 4) & 0x0f) as i8),
-        from_num(((num >> 8) & 0x0f) as i8)
-    ];
-
-    if ret.contains(&'_') {
-        return Err(Error::new(ErrorKind::Other, format!("Invalid orientation data! {} {:#x} {:#b}: {:?}", num, num, num, ret)));
-    }
-
-    // Normalize to common orientation - swap X/Y and invert Z
-    // TODO: I don't think it belongs here
-    ret.swap(0, 1);
-    ret[2] = invert_case(ret[2]);
-
-    Ok(ret.iter().collect())
+    Ok(vec![
+        from_num((num & 0x0f) as i8)?,
+        from_num(((num >> 4) & 0x0f) as i8)?,
+        from_num(((num >> 8) & 0x0f) as i8)?
+    ].iter().collect())
 }
 
-fn read_rational(d: &mut Cursor::<&[u8]>) -> Result<f32> {
+fn read_rational(d: &mut Cursor::<&[u8]>) -> Result<f64> {
     let n = d.read_i32::<BigEndian>()? as f64;
     let d = d.read_i32::<BigEndian>()? as f64;
     if d > 0.0 {
-        Ok((n / d) as f32)
+        Ok(n / d)
     } else {
         Err(Error::new(ErrorKind::Other, "Invalid rational"))
     }
